/************************************************************************* 
 
    Quick And Dirty Apricot Emulator
    Copyright (C) 2005, 2007  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "qdae.hxx"
#include "qdae_ui.hxx"
#include "sdl_video.hxx"
#include "f1_video.hxx"
#include "SdlContext.hxx"
#include "SdlTerm.hxx"
#include "sdl_keyboard.hxx"
#include "wd2797.hxx"

#define SCREEN_W 640
#define SCREEN_H 256
#define SCREEN_DEPTH 32
#define SCREEN_FLAGS 0
#define SCREEN_EXTRA 16
#define FPS_DELAY 100	/* Giving 10fps */
#define VRAMBASE 0x800
#define VRAMLEN  0x1000
#define PALETTE 0xE0000
#define VPTRBASE 0x1E00
#define VPTRLEN  0x1000

F1_Video::F1_Video(PC_ENV *m) : Video(m, "f1", SCREEN_W, SCREEN_H * 2 + SCREEN_EXTRA)
{
	extra_height = SCREEN_EXTRA;
	vscale = 2;
	scr_rows = 256;
}

F1_Video::~F1_Video()
{
}

void F1_Video::write(u_int16 port, u_int8 value)
{
	switch(port)
	{
		case 13: /* Set Stop LED */
			gl_stop = value;
			gl_ledsdirty = 1;
			break;

		case 11: /* Set Caps Lock LED */
			gl_caps = value;
			gl_ledsdirty = 1;
			break;

		case 9: /* Set horizontal resolution */
			v_mode = value;
			gl_dirty = 1;
			break;

		case 7: /* Set vertical resolution */
			gl_dirty = 1;
			if (value == 0) 
			{
				SDL_Rect rc;
	
				scr_rows = 200;
				rc.x = 0;
				rc.y = 200 * vscale;
				rc.w = SCREEN_W;
				rc.h = (SCREEN_H - 200) * vscale;
				SDL_FillRect(vidSurface, &rc, colours[15]);
			}
			else	scr_rows = 256;
			break;
	}
}

void F1_Video::draw_row_2bit(int y, unsigned char *row, unsigned char *shadow,
			unsigned char *palette)
{
	int y1, x1, x;
	unsigned char mask;
	unsigned char *font;
	unsigned char *framebuf;
	unsigned offset;
	unsigned char *bitmap;
	unsigned bpp;
	int col;

	bpp = vidSurface->format->BytesPerPixel;
       	framebuf = (unsigned char *)(vidSurface->pixels);
	for (y1 = 0; y1 < vscale; y1++)
	{
		offset = ((y * vscale + y1) * vidSurface->pitch);

		for (x = 0; x < 160; x += 2, offset += 8 * bpp)
		{
			if (shadow && (row[x] == shadow[x]) &&
					(row[x+1] == shadow[x+1])) continue;
			mask = 0x80;
			for (x1 = 0; x1 < 8; x1++)
			{
				u_int8 d0 = row[x]   & mask;
				u_int8 d1 = row[x+1] & mask;
				if (d0 && d1) col = 10;
				else if (d1)  col = 8;
				else if (d0)  col = 2;
				else          col = 0;
				*(Uint32 *)(&framebuf[offset + x1 * bpp]) = 
					colours[palette[col*2] & 0xF];
				mask = mask >> 1;
			}
		}
	}
}	



void F1_Video::draw_row_4bit(int y, unsigned char *row, unsigned char *shadow,
			unsigned char *palette)
{
	int y1, x1, x;
	unsigned char mask;
	unsigned char *font;
	unsigned char *framebuf;
	unsigned offset;
	unsigned char *bitmap;
	unsigned bpp;
	int col;

	bpp = vidSurface->format->BytesPerPixel;
       	framebuf = (unsigned char *)(vidSurface->pixels);
	for (y1 = 0; y1 < vscale; y1++)
	{
		offset = ((y * vscale + y1) * vidSurface->pitch);

		for (x = 0; x < 160; x += 2, offset += 8 * bpp)
		{
			if (shadow && (row[x] == shadow[x]) &&
					(row[x+1] == shadow[x+1])) continue;
			mask = 0xC0;
			for (x1 = 0; x1 < 4; x1++)
			{
				u_int8 d0 = row[x]   & mask;
				u_int8 d1 = row[x+1] & mask;

				switch(x1)
				{
					case 0: col = (d0 >> 6) | (d1 >> 4); 
						break;
					case 1: col = (d0 >> 4) | (d1 >> 2); 
						break;
					case 2: col = (d0 >> 2) | (d1); 
						break;
					case 3: col = (d1 << 2) | (d0);
						break;
				}
				*(Uint32 *)(&framebuf[offset + (2*x1) * bpp]) = 
					colours[palette[col*2] & 0xF];
				*(Uint32 *)(&framebuf[offset + (2*x1+1)*bpp]) = 
					colours[palette[col*2] & 0xF];
				mask = mask >> 2;
			}
		}
	}
}	

void F1_Video::draw_row(int y, unsigned char *row, unsigned char *shadow,
			unsigned char *palette)
{
	if (v_mode) draw_row_2bit(y, row, shadow, palette);
	else	    draw_row_4bit(y, row, shadow, palette);
}


int F1_Video::paint()
{
	int y;
	int any;
	long offset, shadow_offset;
	SDL_Event e;
	gl_dirty = 1;
	static unsigned char shadow_palette[32];

	for (;;)
	{
		SDL_Delay(FPS_DELAY);

		// If emulator UI is using the screen, don't paint.
		if (gl_ui->usingScreen()) continue;

		if (memcmp(sysram + PALETTE, shadow_palette, 32))
		{
			gl_dirty = 1;
			memcpy(shadow_palette, sysram + PALETTE, 32);	
/*			printf("New palette: ");
			for (y = 0; y < 32; y+=2)
			{
				if (shadow_palette[y] != sysram[y + PALETTE])
				{
					printf("<%02x>", sysram[y + PALETTE]);

				}
				else printf(" %02x ", sysram[y + PALETTE]);
			}
			printf("\n"); */
		}
		any = 0;
		for (y = 0; y < scr_rows /* 256 */; y++)
		{
			offset        = 2 * sysram   [VPTRBASE + 2 * y] + 512 * sysram   [VPTRBASE + 2 * y + 1];
			shadow_offset = 2 * shadowram[VPTRBASE + 2 * y] + 512 * shadowram[VPTRBASE + 2 * y + 1];
			if ((offset != shadow_offset) || gl_dirty)
			{
				if (!any) SDL_LockSurface(vidSurface);
				any = 1;
				draw_row(y, sysram + offset, NULL, sysram + PALETTE);
			}
			else if (memcmp(sysram + offset, shadowram + offset, 160))
			{
				if (!any) SDL_LockSurface(vidSurface);
				any = 1;
				draw_row(y, sysram + offset, 
						/*NULL */ shadowram + offset, 
						sysram + PALETTE);
				memcpy(shadowram + offset, sysram + offset, 
						160);
			}
			/* else  // Force drawing
			{
				if (!any) SDL_LockSurface(vidSurface);
				any = 1;
				draw_row(y, sysram + offset, NULL, sysram + PALETTE);
			} */
		}
		if (any)
		{
			memcpy(shadowram + VPTRBASE, sysram + VPTRBASE, 512);
			SDL_UpdateRect(vidSurface, 0, 0, 0, 0);
			SDL_UnlockSurface(vidSurface);
		}
		if (gl_dirty || gl_ledsdirty)
		{
			term->setForeground(0, 0xFF, 0);
			term->moveCursor(0, 0);
			(*term) << (gl_caps ? "CAPS " : "     ");
			term->setForeground(0xFF, 0xFF, 0);
			(*term) << (gl_stop ? "STOP " : "     ");
			term->setForeground(0xFF, 0x00, 0);
			(*term) << (floppy_indicator[0] ? "A " : "  ");
			(*term) << (floppy_indicator[1] ? "B " : "  ");
			gl_ledsdirty = 0;
		}
		gl_dirty = 0;
	}
}

/* Fairly arbitrary 16-colour palette, in the same order as the
 * Rascal C (F-series)  */
static int r[16] = {  255, 191,   0,   0, 255, 191,   0,   0,
		      255, 191,   0,   0, 255, 191,  63,   0};

static int g[16] = {  255, 191, 255, 191,   0,   0,   0,   0,
		      255, 191, 255, 191,   0,   0,  63,   0 };

static int b[16] = {  255, 191, 255, 191, 255, 191, 255, 191,
     		        0,   0,   0,   0,   0,   0,  63,   0};


void F1_Video::init()
{
	int n;

	Video::init();

	sysram = host->m_mem_base;
	memcpy(shadowram, sysram, 0x4000);
	for (n = 0; n < 16; n++)
	{
		colours[n] = SDL_MapRGB(vidSurface->format, r[n], g[n], b[n]);
	}
}
