/*************************************************************************
 
    Quick And Dirty Apricot Emulator
    Copyright (C) 2007  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "qdae.hxx"
#include "qdae_ui.hxx"
#include "device.hxx"
#include "i8089.hxx"

#include "UiContainer.hxx"
#include "UiLabel.hxx"
#include "UiSeparator.hxx"
#include "UiSetting.hxx"
#include "UiCommand.hxx"
#include "UiMenu.hxx"

#define MEMORY_TO_MEMORY 0xC008	/* Source is GA */
#define DISK_TO_MEMORY   0x8848	/* Source is GA */
#define MEMORY_TO_DISK	 0x5048	/* Source is GA */
#define MEMORY_TO_DISK2	 0x5448	/* Source is GB */

			/* Right. That gives the meaning of one bit in this
			 * register: bit 10 is 0 for GA source, 
			 * 1 for GB source */

/* 8089 system info starts at FFFF6h with an info block:
 * 
 * 	DW 0FF01h	;16-bit bus
 * 	DD 0		;Control block at 0:0
 *
 * 	The first time it's started up it therefore looks at 0:0 for its
 * 	system configuration info.
 *
 * 8089 asm:
 *
 * 	ppp0 0aa1  1000 10mm	LPD p, m
 *
 */
//#define I89_LOG(x) printf x ; fflush(stdout);
#define I89_LOG(x)

I8089::I8089(PC_ENV *host) : Device("i8089")
{
	m_inited = false;
	m_host = host;
	m_ch[0].m_parent = m_ch[1].m_parent = this;
	m_ch[0].m_host   = m_ch[1].m_host   = host;
	m_ch[0].m_offset = 0;
	m_ch[1].m_offset = 8;
	reset();

//	enable(false);	// XXX Add menu via f6 to enable/disable.
}

I8089::~I8089()
{

}

void I8089::write(u_int16 port, u_int16 value)
{
/* The 8089's RESET signal is the same as the 8086's.
 *
 * Its SEL signal comes from line A1, which appears to be buffered from the
 * processor's AD1 line. I suspect it would therefore track writes to 
 * port 72h and writes of 02h to port 70h.
 *
 * Its CA signal is active for any I/O writes in the 70h-77h range.
 * 		
 * */
	I89_LOG(("8089: OUT %d, %d\n", port, value));

	if (!isEnabled()) return;

	if (port == 0)
	{
// Assume any other writes are just to stop the 8089 after it's finished.
		if (value != 0)
		{
//			printf("XXX What does OUT %d, %d mean?\n", port, value);
			return;
		}
		m_ch[0].execute();
	}
	if (port == 1 && !m_inited)
	{
		initialise();
		m_inited = true;
	}
}

u_int16 I8089::read(u_int16 port)
{
	u_int16 value = 0xFFFF;

	I89_LOG(("8089: IN %d = %d\n", port, value));
}


void I8089::reset()
{
	m_inited = false;
	m_ch[0].reset();
	m_ch[1].reset();
}


bool I8089::parseNode(xmlDoc *doc, xmlNs *ns, xmlNode *node)
{
	return parseEnabled(doc, ns, node);
}

bool I8089::storeNode(xmlDoc *doc, xmlNs *ns, xmlNode *node)
{
	return storeEnabled(doc, ns, node);
}

bool I8089::hasSettings(SDLKey *key, string &caption)
{
	*key = SDLK_8;
	caption = "  8089 I/O processor  ";
	return true;
}

UiEvent I8089::settings(UiDrawer *d)
{
	int sel = -1;
	UiEvent e;

        while (1)
        {
                UiMenu uim(d);
                bool bEnabled;

                bEnabled = isEnabled();
                uim.add(new UiLabel("                 8089 I/O Processor", d));
                uim.add(new UiSeparator(d));
				/*   1...5...10...15...20...25...30...35...40...45...50...55...60 */
		uim.add(new UiLabel("Note: Although the emulated system will pass its self-test", d));
		uim.add(new UiLabel("with the 8089 disabled, most operating systems other than", d));
		uim.add(new UiLabel("MS-DOS (BIOS 2.4) will hang on booting.", d));
		uim.add(new UiLabel("", d));
                uim.add(new UiSetting(SDLK_d, !bEnabled, "  Disabled  ", d));
                uim.add(new UiSetting(SDLK_e,  bEnabled, "  Enabled  ",  d));
	       	uim.add(new UiSeparator(d));
	        uim.add(new UiCommand(SDLK_ESCAPE, "  EXIT  ", d));
		uim.setSelected(sel);
		d->centreContainer(uim);	
		e = uim.eventLoop();
                if (e == UIE_QUIT) return e;
                sel = uim.getSelected();
                switch(uim.getKey(sel))
		{
			case SDLK_d: enable(false); break;
			case SDLK_e: enable();      break;
			case SDLK_ESCAPE: return UIE_CONTINUE;
			default:	break;
		}
	}
	return UIE_CONTINUE;
}




void I8089::tick()
{
	if (m_ch[0].m_countdown || m_ch[0].m_running) m_ch[0].tick();
	if (m_ch[1].m_countdown || m_ch[1].m_running) m_ch[1].tick();
}

void I8089::initialise()
{
	u_int16 scbseg, scboff;
	int scbaddr;
	u_int8 soc;

	/* Find the SCB */
	m_buswidth = m_host->fetch  (0xFFFF6);
	scboff     = m_host->fetch16(0xFFFF8);
	scbseg     = m_host->fetch16(0xFFFFA);

	I89_LOG(("8089 initialising: SCB @ %04x:%04x, bus width %d\n",
			scbseg, scboff, m_buswidth));
	scbaddr = scbseg * 16 + scboff;

	m_soc      = m_host->fetch  (scbaddr);
	m_cboff    = m_host->fetch16(scbaddr + 2);
	m_cbseg    = m_host->fetch16(scbaddr + 4);
	I89_LOG(("8089 initialising: CB @ %04x:%04x, soc=%02x\n",
			m_cbseg, m_cboff, m_soc));

	m_ch[0].m_countdown = 30;
}




void I8089Channel::reset()
{
	m_running = false;
	m_countdown = 0;
/* The whole m_xferPending thing is a hack. To do things properly it 
 * shouldn't schedule an xfer() for some fixed number of instructions later,
 * but wait until the FDC becomes ready and then do the transfer. 
 *
 * Goodness only knows what it'll need doing to it if I ever get round to
 * implementing Winchester support... */
	m_xferPending = 0;

	m_regptr[0] = &m_ga;
	m_regptr[1] = &m_gb;
	m_regptr[2] = &m_gc;
	m_regptr[3] = &m_bc;
	m_regptr[4] = &m_tp;
	m_regptr[5] = &m_ix;
	m_regptr[6] = &m_cc;
	m_regptr[7] = &m_mc;

	m_baseptr[0] = &m_ga;
	m_baseptr[1] = &m_gb;
	m_baseptr[2] = &m_gc;
	m_baseptr[3] = &m_pp;
}

int I8089Channel::addr()
{
	return m_parent->m_cbseg * 16 + m_parent->m_cboff + m_offset; 
}


// Depending on addressing mode (AA field) get the operand address. 
int I8089Channel::getBase(int *baseptr, int mode)
{
	u_int16 offset;
	
	switch(mode & 3)
	{
		case 0: // The selected pointer contains the operand address.
			return *baseptr;
		case 1: // The operand address is formed by adding an 8-bit,
			// unsigned, offset contained in the instruction to
			// the selected pointer. The contents of the pointer
			// are unchanged.
			offset = m_host->fetch(m_ip++);
			return offset + *baseptr;
		case 2: // The operand address is formed by adding the 
			// contents of the index register to the selected
			// pointer. The contents of the pointer are unchanged.
			return m_ix + *baseptr;
		case 3: // Same as 10 except the index register is post 
			// auto-incremented (by 1 for 8-bit transfer, by 2
			// for 16-bit transfer).
			offset = m_ix;
			// XXX postincrement: 8- or 16- bit?
			m_ix++;
			return offset + *baseptr;
	}
}


// Fetch immediate argument as a 20-bit signed int 
int I8089Channel::fetchImmediate(int wb)
{
	int val = 0;

	wb &= 0x18;
	if (wb == 0x10)
	{
		val = m_host->fetch16(m_ip);
		m_ip += 2;
	}
	else if (wb == 8)
	{
		val = m_host->fetch(m_ip++);
		if (val & 0x80) val |= 0xFF00;
	}
	if (val & 0x8000)
	{
		val |= 0xF0000;
	}
	return val;
}


/* These functions (to access the registers) are written for clarity rather
 * than speed. It doesn't help that half the registers are 20 bits wide and
 * the other half are 16 bits wide. */
void I8089Channel::setReg(int reg, int value)
{
	switch(reg)
	{
		case 0: m_ga = value & 0xFFFFF; break;
		case 1: m_gb = value & 0xFFFFF; break;
		case 2: m_gc = value & 0xFFFFF; break;
		case 3: m_bc = value & 0xFFFF; break;
		case 4: m_tp = value & 0xFFFFF; break;
		case 5: m_ix = value & 0xFFFF; break;
		case 6: m_cc = value & 0xFFFF; break;
		case 7: m_mc = value & 0xFFFF; break;
	}
}

void I8089Channel::adjustReg(int reg, int value)
{
	switch(reg)
	{
		case 0: m_ga = (m_ga + value) & 0xFFFFF; break;
		case 1: m_gb = (m_gb + value) & 0xFFFFF; break;
		case 2: m_gc = (m_gc + value) & 0xFFFFF; break;
		case 3: m_bc += value; break;
		case 4: m_tp = (m_tp + value) & 0xFFFFF; break;
		case 5: m_ix += value; break;
/* CC cannot be incremented or decremented 
 		case 6: m_cc += value; break; */
		case 7: m_mc += value; break;
	}
}



const char *I8089Channel::immediateName(int ih)
{
	static char buf[20];
	int ip = m_ip;
	int val = fetchImmediate(ih);
	m_ip = ip;
	sprintf(buf, "0x%05x : %c", val, (ih & 1) ? 'W' : 'B');
	return buf;
}

const char *I8089Channel::memoryName(int ih, int il)
{
	static char buf[30];

	switch((ih >> 1) & 3)
	{
		case 0: sprintf(buf, "[%s]", basePtrName(il)); break;
		case 1: sprintf(buf, "[%s+0x%02x]", basePtrName(il),
					m_host->fetch(m_ip)); break;
		case 2: sprintf(buf, "[%s+IX]", basePtrName(il)); break;
		case 3: sprintf(buf, "[%s+IX++]", basePtrName(il)); break;
	}
	strcat(buf, (ih & 1) ? ": W" : ": B");
	return buf;
}


const char *I8089Channel::rrrName(int reg)
{
	static const char *names[] = 
		{"GA", "GB", "GC", "BC", "TP", "IX", "CC", "MC"};

	return names[reg & 7];
}


const char *I8089Channel::pppName(int reg)
{
	static const char *names[] = 
		{"GA", "GB", "GC", "!3", "TP", "!5", "!6", "!7"};

	return names[reg & 7];
}


const char *I8089Channel::basePtrName(int reg)
{
	static const char *names[] = 
		{"GA", "GB", "GC", "PP"};

	return names[reg & 3];
}



void I8089Channel::tick()
{
	int n;

	if (m_countdown)
	{
/* If finishing up an operation, reset busy flag */
		--m_countdown;
		if (m_countdown == 0)
		{
			m_host->store(addr() + 1, 0);
		}
	}
/* If running execute an instruction */
	if (m_running)
	{
		int *ptr;
		int16 *ptr16;
		int16 val, val2;
		int *baseptr;
		int base;
		int wb, reg;

		u_int8 ih, il;
		u_int16 instr;
		I89_LOG(("8089 Status:  GA=%05x GB=%05x GC=%05x BC=%05x IP=%05x\n"
			 "              TP=%05x IX=%05x CC=%05x MC=%05x\n",
			m_ga, m_gb, m_gc, m_bc, m_ip, m_tp, m_ix, m_cc, m_mc));
/* Note big-endian instruction */
		ih = m_host->fetch(m_ip++);
		il = m_host->fetch(m_ip++);
		instr = ih * 256 + il;

/* Decoding... */
		wb = ih & 0x18;
		reg = (ih >> 5) & 7;
		I89_LOG(("8089 Decode:  %02x %02x  wb=%d reg=%d  ", 
				ih, il, (wb >> 3), reg));
		switch(il & 0xFC)
		{
			case 0x00: 
				switch(ih)
				{
					case 0: /* NOP */
					I89_LOG(("NOP\n"));
					break;
					case 0x80: case 0xA0: 
					case 0xC0: case 0xE0: /* WID */
					m_sourcew = (ih & 0x40) ? 16 : 8;
					m_destw   = (ih & 0x20) ? 16 : 8;
					I89_LOG(("WID %d, %d\n", m_sourcew, m_destw));
					break;
					
					case 0x60: /* XFER */
					I89_LOG(("XFER\n"));

/* The CP/M-86 8089 code has this thing where it does XFER immediately 
 * _followed_ by the I/O operation to set the transfer going (whereas the
 * ROM BIOS does these the other way round). 
 *
 * Quick workaround: Delay an XFER by one instruction. */
					m_xferPending = 1;
					break;
/*
					case 0x40:
					I89_LOG(("SINTR\n"));	// XXX
					break;
*/
					default:	/* Unknown */
{
	char buf[80];

	sprintf(buf, "Unimplemented opcode %02x %02x", ih, il);
	if (gl_ui->i8089_error(buf) == UIE_QUIT) exit(1);
	m_countdown = 12;
	m_running = 0;
}
					break;
				}
				break;
/*			case 0x08: 
				I89_LOG(("LPDI P,I\n"));	// XXX
				break;
*/
			case 0x20: /* JMP */
				I89_LOG(("JMP %s\n", immediateName(ih)));
				val = fetchImmediate(wb);
				m_ip += (signed short)val;
				break;
			case 0x30: /* MOVI */
				I89_LOG(("MOVI %s,%s\n", rrrName(reg),
						immediateName(ih)));
				val   = fetchImmediate(wb);
				setReg(reg, val);
				break;
			case 0x38: /* INC */
				I89_LOG(("INC %s\n", rrrName(reg)));
				adjustReg(reg, 1);
				break;
			case 0x3C: /* DEC */
				I89_LOG(("DEC %s\n", rrrName(reg)));
				adjustReg(reg, -1);
				break;
			case 0x40: /* JNZ R */
				I89_LOG(("JNZ %s,%s\n", rrrName(reg), 
						immediateName(ih)));
				val = fetchImmediate(wb);
				ptr16 = (int16 *)(m_regptr[reg]);
				if (*ptr16)
				{
					m_ip += (signed short)val;
				}
				break;
			case 0x44: /* JZ R */
				I89_LOG(("JZ %s,%s\n", rrrName(reg), 
						immediateName(ih)));
				val = fetchImmediate(wb);
				ptr16 = (int16 *)(m_regptr[reg]);
				if (!*ptr16)
				{
					m_ip += (signed short)val;
				}
				break;
			case 0x48: /* HLT */
				I89_LOG(("HLT\n"));
				m_countdown = 12;
				m_running = 0;
				break;
			case 0x4C: /* MOVI M, */
				I89_LOG(("MOVI %s,", memoryName(ih, il)));
				baseptr = m_baseptr[il & 3];
				base = getBase(baseptr, (ih >> 1) & 3);
				I89_LOG(("%s\n", immediateName(ih)));
				val = fetchImmediate(wb);
				if (ih & 1)
					store16(il, base, val);
				else	store8 (il, base, val);
				break;
			case 0x80: /* MOV R, M */
				I89_LOG(("MOV %s,%s\n",
					rrrName(reg), memoryName(ih, il)));
				baseptr = m_baseptr[il & 3];
				base = getBase(baseptr, (ih >> 1) & 3);
				val = m_host->fetch16(base); 
				setReg(reg, val);
				break;
			case 0x84: /* MOV M, R */
				I89_LOG(("MOV %s,%s\n",
					memoryName(ih, il), rrrName(reg)));
				ptr16 = (int16 *)(m_regptr[reg]);
				baseptr = m_baseptr[il & 3];
				base = getBase(baseptr, (ih >> 1) & 3);
				if (ih & 1)
					store16(il, base, *ptr16); 
				else	store8 (il, base, *ptr16);
				break;
			case 0x88: /* LPD P, M */
				I89_LOG(("LPD %s,%s\n", pppName(reg),
						memoryName(ih, il)));
				ptr = (int *)(m_regptr[reg]);
				baseptr = m_baseptr[il & 3];
				base = getBase(baseptr, (ih >> 1) & 3);
				*ptr = m_host->fetch20(base);
				break;
/*
			case 0x8C:
				I89_LOG(("MOVP P,M\n"));
				break;
*/
			case 0x90:
				I89_LOG(("MOV M,M (source=%s) ",
					memoryName(ih, il)));
				baseptr = m_baseptr[il & 3];
				base = getBase(baseptr, (ih >> 1) & 3);
				if (ih & 1)
					m_mmcopy = fetch16(il, base);
				else	m_mmcopy = fetch8 (il, base);
				I89_LOG((" (%04x)\n", m_mmcopy));
/* Treat the two halves of MOV M,M as a single instruction */
				if (m_xferPending) --m_xferPending;
				break;
/*
			case 0x98: 
				I89_LOG(("MOVP M,P\n"));
				break;
*/
			case 0xBC: /* JBT */
				I89_LOG(("JBT %d,%s\n", reg, immediateName(ih),
							memoryName(ih,il)));
				baseptr = m_baseptr[il & 3];
				base = getBase(baseptr, (ih >> 1) & 3);
				val = fetch8 (il, base);
/* I think this is right: jump displacement comes last */
				val2 = fetchImmediate(wb);
				if (val & (1 << reg))	
				{
					m_ip += (signed short)val2;
				}
				break;
			case 0xCC:
				baseptr = m_baseptr[il & 3];
				base = getBase(baseptr, (ih >> 1) & 3);
				I89_LOG(("MOV M,M (dest=%s)\n",
					memoryName(ih, il)));
				if (ih & 1)
					store16(il, base, m_mmcopy);
				else	store8 (il, base, m_mmcopy);
			        break;	
			default:
				/* XXX Treat unknown ops like HLT */
{
	char buf[80];

	sprintf(buf, "Unimplemented opcode %02x %02x", ih, il);
	if (gl_ui->i8089_error(buf) == UIE_QUIT) exit(1);
	m_countdown = 12;
	m_running = 0;
}
				break;
		}
		if (m_xferPending)
		{
			++m_xferPending;
			if (m_xferPending == 3)
			{
				m_xferPending = 0;
				xfer();
			}
		}
	}
}


void I8089Channel::execute()
{
	I89_LOG(("8089 Execute: addr=%05x\n", addr()));

	m_ccw   = m_host->fetch  (addr());
	m_pboff = m_host->fetch16(addr() + 2);
	m_pbseg = m_host->fetch16(addr() + 4);

	m_tboff = m_host->fetch16(m_pbseg, m_pboff);
	m_tbseg = m_host->fetch16(m_pbseg, m_pboff + 2);

	I89_LOG(("              PB @ %04x:%04x, ccw=%02x\n",
			m_pbseg, m_pboff, m_ccw));
	I89_LOG(("              TB @ %04x:%04x\n", m_tbseg, m_tboff));

	m_ga = m_gb = m_gc = 0;
	m_tp = m_tbseg * 16 + m_tboff;
	m_pp = m_pbseg * 16 + m_pboff;
	m_ip = m_tp;
	m_running = true;
}

u_int8 I8089Channel::fetch8(int base, int addr)
{
	switch(base & 3)
	{
		case 0: /* GA */ return m_host->fetch(addr);
		case 1: /* GB */ return m_host->fetch(addr);
		case 2: /* GC */ if (m_cc == DISK_TO_MEMORY || 
				     m_cc == MEMORY_TO_DISK ||
				     m_cc == MEMORY_TO_DISK2)
					return m_host->sys_in8(addr);
				 else	return m_host->fetch(addr);
		case 3: /* PP */ return m_host->fetch(addr);
	}	
}


void I8089Channel::store8(int base, int addr, u_int8 value)
{
	switch(base & 3)
	{
		case 0: /* GA */ m_host->store(addr, value);
				 break;
		case 1: /* GB */ m_host->store(addr, value);
				 break;
		case 2: /* GC */ if (m_cc == DISK_TO_MEMORY ||
				     m_cc == MEMORY_TO_DISK ||
				     m_cc == MEMORY_TO_DISK2)
				 	m_host->sys_out8(addr, value);
				 else	m_host->store(addr, value);
				 break;
		case 3: /* TP */ m_host->store(addr, value);
				 break;
	}	
}

u_int16 I8089Channel::fetch16(int base, int addr)
{
	switch(base & 3)
	{
		case 0: /* GA */ return m_host->fetch16(addr);
		case 1: /* GB */ return m_host->fetch16(addr);
		case 2: /* GC */ if (m_cc == DISK_TO_MEMORY ||
				     m_cc == MEMORY_TO_DISK ||
				     m_cc == MEMORY_TO_DISK2)
					return m_host->sys_in16(addr);
				 else	return m_host->fetch16(addr);
		case 3: /* PP */ return m_host->fetch16(addr);
	}	
}


void I8089Channel::store16(int base, int addr, u_int16 value)
{
	switch(base & 3)
	{
		case 0: /* GA */ m_host->store16(addr, value);
				 break;
		case 1: /* GB */ m_host->store16(addr, value);
				 break;
		case 2: /* GC */ if (m_cc == DISK_TO_MEMORY ||
				     m_cc == MEMORY_TO_DISK ||
				     m_cc == MEMORY_TO_DISK2)
				 	m_host->sys_out16(addr, value);
				 else	m_host->store(addr, value);
				 break;
		case 3: /* PP */ m_host->store16(addr, value);
				 break;
	}
}


void I8089Channel::xfer()
{
	int n;
	I89_LOG(("\n8089 XFER>>: ga=%05x gb=%05x bc=%04x cc=%04x <<\n\n",
				m_ga, m_gb, m_bc, m_cc));

	/* Known values of m_cc:
	 * 0xC008: memory-to-memory DMA, from ga   to gb
	 * 0x8848: disk-to-memory DMA,   from disk to gb
	 * 0x5048: disk-to-memory DMA,   from ga to disk
	 * 0x5448: disk-to-memory DMA,   from gb to disk
	 */

	/* XXX This assumes that:
	 * - the operation is always 8 bits wide
	 * - the 8089 doesn't do anything weird with the data en route 
	 */
	switch(m_cc)
	{
		case MEMORY_TO_MEMORY:
			for (n = 0; n < m_bc; n++)
			{
				store8(1, m_gb++, fetch8(0, m_ga++));
			}
			break;
		case DISK_TO_MEMORY:
			for (n = 0; n < m_bc; n++)
			{
/* I don't know where the '+6' comes from but it seems to be necessary for 
 * this to work. Probably in a real system it doesn't use an I/O address 
 * at all, just asks the floppy controller for DMA transfers. */
				u_int8 val = fetch8(2, m_gc+6);
				store8(1, m_gb++, val);
//// Dump the bytes read from the disc.
////				printf("%02x ", val);
////				if ((n % 16) == 15) putchar('\n');
//				store8(1, m_gb++, fetch8(2, m_gc + 6));
			}
			break;
		case MEMORY_TO_DISK:
			for (n = 0; n < m_bc; n++)
			{
/* I don't know where the '+6' comes from but it seems to be necessary for 
 * this to work. Probably in a real system it doesn't use an I/O address 
 * at all, just asks the floppy controller for DMA transfers. */
				u_int8 val = fetch8(1, m_ga++);
				store8(2, m_gc+6, val);
//				store8(2, m_gc+6, fetch8(1, m_ga++));
			}
			break;
		case MEMORY_TO_DISK2:
			for (n = 0; n < m_bc; n++)
			{
/* I don't know where the '+6' comes from but it seems to be necessary for 
 * this to work. Probably in a real system it doesn't use an I/O address 
 * at all, just asks the floppy controller for DMA transfers. */
				u_int8 val = fetch8(1, m_gb++);
				store8(2, m_gc+6, val);
//				store8(2, m_gc+6, fetch8(1, m_ga++));
			}
			break;
		default:
		{
			char buf[80];

			sprintf(buf, "Unimplemented transfer mode 0x%04x", m_cc);
			if (gl_ui->i8089_error(buf) == UIE_QUIT) exit(1);
			m_countdown = 12;
			m_running = 0;
		}
			break;
	}
}
