/*************************************************************************
 
    Quick And Dirty Apricot Emulator
    Copyright (C) 2007  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "qdae.hxx"
#include "device.hxx"
#include "i8259.hxx"

//#define PIC_LOG(x) printf x ; fflush(stdout);
#define PIC_LOG(x)

I8259::I8259(PC_ENV *host)
{
	m_host = host;
	reset();
}


I8259::~I8259()
{

}

void I8259::setICW1(u_int8 value)
{
	m_icw1 = value;
	m_haveIcw2 = true;
	m_haveIcw3 = (value & 2) == 0;
	m_haveIcw4 = (value & 1) != 0;
	PIC_LOG(("PIC: icw1=%02x\n", m_icw1));
}


void I8259::write(u_int16 port, u_int8 value)
{
	PIC_LOG(("PIC: Out %d, %02x\n", port, value));
	switch(port)
	{
		case 0:	if (m_haveOcw2)
			{
				m_ocw2 = value;
				m_haveOcw2 = false;
				break;
			}
			if (m_haveOcw3)
			{
				m_ocw3 = value;
				m_haveOcw3 = false;
				break;
			}
			if (value & 0x10)
			{
				setICW1(value);
				break;
			}
			if (value & 0x20) // EOI
			{
				int irq;
				u_int8 mask = 1;

/* Acknowledge specific interrupt */
				if (value & 0x40) 
				{
					irq = value & 7;
					mask = 1 << irq;
				}
/* Acknowledge highest-priority interrupt */
				else for (irq = 0; irq < 8; irq++)
				{
					if (m_isr & mask) break;
					mask = mask << 1;
				}
				PIC_LOG(("PIC: EOI for IRQ %d\n", irq));

//		if (irq == 6 || irq == 5) m_host->settings.debug &= ~(DEBUG_DECODE_F | DEBUG_TRACE_F);
				if (irq < 8) 
				{
					m_isr &= ~mask;
				}
				break;
			}
			PIC_LOG(("PIC: Unknown command %02x\n", value));
			break;
		case 1:	if (m_haveIcw2)
			{
				m_icw2 = value;
				m_haveIcw2 = false;
				return;	
			}
			if (m_haveIcw3)
			{
				m_icw3 = value;
				m_haveIcw3 = false;
				return;	
			}
			if (m_haveIcw4)
			{
				m_icw4 = value;
				m_haveIcw4 = false;
				return;	
			}
			setMask(value);
			break;
	}
}


u_int8 I8259::read(u_int16 port)
{
	u_int8 value = 0xFF;

	switch(port)
	{
		case 0: value = (m_ocw3 & 1) ? m_isr : m_irr; break;
		case 1: value = m_mask;   break;
	}
	PIC_LOG(("PIC: Read %d = %02x\n", port, value));
	return value;
}

void I8259::setMask(u_int8 value)
{
	int irq;
	int match = 0x01;
	bool stillRaising = false;
	bool stopped = false;

//
// We are masking off an interrupt. There may be an interrupt of this type
// pending; if so, squash it. If there are no other interrupts pending, then
// stop bothering the CPU.
//
	PIC_LOG(("PIC: Set mask to %02x\n", value));
	m_mask = value;
	for (irq = 0; irq < 8; irq++, match = match << 1)
	{
		if (m_mask & match)
		{
			if (m_raising[irq])
			{
				PIC_LOG(("PIC: Clear pending interrupt on IRQ%d\n", irq));
				m_raising[irq] = false;
				stopped = true;
			}
		}
		else	// Still raising something that isn't masked?
		{
			if (m_raising[irq]) stillRaising = true;
		}
	}
	if (stopped && !stillRaising)
	{
		m_host->i86_intr_lower(this);
	}
}

void I8259::reset()
{
	m_icw1 = m_icw2 = m_icw3 = m_icw4 = 0;
	m_haveIcw2 = m_haveIcw3 = m_haveIcw4 = false;
	m_haveOcw2 = m_haveOcw3 = false;
	m_irr = m_isr = 0;
	memset(m_lastirq, 0, sizeof(m_lastirq));
	memset(m_realirq, 0, sizeof(m_realirq));
	memset(m_raising, 0, sizeof(m_raising));
	setMask(0xFF);
}


bool I8259::parseNode(xmlDoc *doc, xmlNs *ns, xmlNode *node)
{
	return true;
}


bool I8259::storeNode(xmlDoc *doc, xmlNs *ns, xmlNode *node)
{
	return true;
}

void I8259::tick()
{
	int irq;
	int match = 1;

/* Check for level-triggered interrupts that are now unmasked */
	if ((m_icw1 & 8) == 0)	// If edge-triggered, missed interrupts stay
		return;		// missed.

	// XXX Don't allow any interrupt to override any other.
	if (m_isr) return;

	for (irq = 0; irq < 8; irq++, match = match << 1)
	{
		/* An interrupt is being handled at this level. 
		 * No lower- or same- priority interrupt can intervene. */
		if (m_isr & match) break;
		/* If an IRQ is high... */
		if (!m_realirq[irq]) continue;
		/* And it isn't masked... */
		if (m_mask & match) continue;	
		/* Then deal with it */
		setirq(irq, m_realirq[irq]);
	}
}

void I8259::setirq(int irq, bool on)
{
	int match = (1 << irq);

	PIC_LOG(("PIC: IRQ%d -> %d\n", irq, on));
	m_realirq[irq] = on;
	/* Deal with edge-triggered vs level-triggered */
	if ((m_icw1 & 8) != 0)	// level triggered
	{
		m_lastirq[irq] = on; 
		if (!on) return;	/* Interrupt has gone low */
	}
	else			// edge triggered
	{
		if (m_lastirq[irq] == on) return;	// no edge
		m_lastirq[irq] = on;
		if (!on) return;		// falling edge
	}
	if (m_mask & match)
	{
		PIC_LOG(("PIC: Suppress IRQ%d\n", irq));
		return;	/* Masked off */
	}
	if (m_isr & match)
	{
		PIC_LOG(("PIC: Ignore IRQ%d, already in its handler\n", irq));
		return;	/* Masked off */
	}
	/* OK, we're dealing with an interrupt */
	m_raising[irq] = true;
	PIC_LOG(("PIC: Interrupt CPU, irq=%d\n", irq));

	m_host->i86_intr_raise(this);
//	i86_intr_raise(m_host, irq + m_icw2);
}

u_int8 I8259::ackInterrupt()
{
	u_int8 match = 1;
	int irq, j;
	char s[9];
	/* Find the highest priority unmasked IRQ */

	for (j = 0; j < 8; j++) {s[j] = m_raising[j] + '0'; } s[j] = 0;
	PIC_LOG(("PIC [%s]: CPU acknowledges interrupt\n", s, irq + m_icw2));
	for (irq = 0; irq < 8; irq++, match = match << 1)
	{
		if (m_mask & match) continue;	/* Masked off */
		if (m_isr & match) continue;	/* Already being processed */
		if (m_raising[irq]) break;
	}
	if (irq == 8)
	{
		fprintf(stderr, "PIC: Received INTA when nothing "
				"was interrupting.\n");
		return 3;	/* Debug break */
	}
	m_raising[irq] = 0;
	m_isr |= match;

//	if (irq + m_icw2 == 0x56 || irq + m_icw2 == 0x55)
//	{
//		m_host->settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
//	}
	for (j = 0; j < 8; j++) {s[j] = m_raising[j] + '0'; } s[j] = 0;
	PIC_LOG(("PIC [%s]: INTA returns 0x%02x\n", s, irq + m_icw2));
	return irq + m_icw2;
}
