
/* 
 * Dos/PC Emulator
 * Copyright (C) 1991 Jim Hudgens
 * 
 * 
 * The file is part of GDE.
 * 
 * GDE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * GDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GDE; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 *
 */

static char rcsid[]=
  "$Id: $";

/* $Log:  $
 * Revision 0.02  1991/07/30  02:09:56  hudgens
 * added copyright.
 *
 * Revision 0.01  1991/07/20  04:10:56  hudgens
 * Initial revision
 *
 */



#include "qdae.hxx"
#include "qdae_files.hxx"
#include "qdae_ui.hxx"
#include "sdl_keyboard.hxx"
#include "sdl_video.hxx"
#include "memsize.hxx"
#include "speed.hxx"

#include "UiContainer.hxx"
#include "UiLabel.hxx"
#include "UiSeparator.hxx"
#include "UiCommand.hxx"
#include "UiMenu.hxx"

void debug_init(PC_ENV *m);

PC_ENV::PC_ENV(const char *hwfile, const char *xmlns)
{
	m_keyboard = NULL;
	m_video = NULL;
	m_cputicks = 0;
	m_devices = new Device *[10];
	m_devCount = 0;
	m_devMax = 10;
	m_hwfile = hwfile;
	m_xmlns  = xmlns;
	m_memsize = new Memsize();
	m_speed = new Speed();
	addDevice(m_memsize);
	m_settings = xmlNewDoc((xmlChar *)"1.0");
	enc_str_pos = 0;
	enc_pos = 0;
        if (m_settings)      
        {                    
                xmlNsPtr ns; 
                xmlNodePtr root;

                root = xmlNewNode(NULL, (xmlChar *)"qdae");
                xmlDocSetRootElement(m_settings, root);

                ns = xmlNewNs(root, (xmlChar *)m_xmlns, NULL);
        }
	m_timers = 0;
	m_ticks = SDL_GetTicks();
	m_delay = 0;
}

PC_ENV::~PC_ENV()
{
	delete m_speed;
	delete m_memsize;
	delete [] m_devices;
}

/** 
 **  do the following:
 **   allocate all the memory to be used.  Currently 1M.
 **   set up interrupt table.
 **
 */
  
void PC_ENV::initialize()
{
	int i;

	addDevice(m_speed);
	/* start up the debugging stuff */
	/* needs to be done *always* whether DEBUG defined or not */
	debug_init();

	/* initialize system memory. Whatever size the user has asked for, we
	 * always allocate 1M so the BIOS and video hardware can live at the 
	 * top */

	m_mem_base = (u_int8 *) malloc(1024 * 1024);

	if (! m_mem_base) 
	{
		sys_fatal(errno, "malloc of system memory sized %d failed",
				1024 * 1024);
	}
	m_video->init();
}

void PC_ENV::reboot()
{
	R_CS = 0xFFFF;
	R_IP = 0;
	/* Zap memory */
	memset(m_mem_base, 0, 0xF0000);
	for (int n = 0; n < m_devCount; n++)
	{
		m_devices[n]->reset();
	}
}


void PC_ENV::clock_tick()
{
/* Poll for events every 10000 instruction fetches. */
	++m_cputicks;
	if (m_cputicks == 10000) 
	{
		m_cputicks = 0;
		poll_events();
	}
	// Speed regulation disabled if speed == 0
	if (m_speed->m_speed)
	{
		if (R_CS < 0xF800) ++m_timers;
		/* Rather brutal speed regulation. If QDAE is running fast 
		 * enough that timer interrupts are closer together than 
		 * 20ms, then space them out with SDL_Delay. 
		 *
		 * Speed regulation has been disabled if we're in the BIOS 
		 * code segment, so that the memory test runs at the full 
		 * speed. Naughty.
		 */
		if (m_timers >= m_speed->m_speed)
		{
			long t = SDL_GetTicks();
			long delta = t - m_ticks;
	
			if (delta >= 0 && delta < 20)	
				/* ie, if interrupts are closer to 20ms apart */
			{
				m_delay += (20 - delta);
			}
/* If we get more than TIMER_RESOLUTION ms adrift, attempt to compensate.
 * Allow it to build up to 3 * TIMER_RESOLUTION so that we can deal with
 * the overhead of the delay. */
			if (m_delay >= 3 * TIMER_RESOLUTION)
			{
				long wait = (m_delay / TIMER_RESOLUTION) 
						 * TIMER_RESOLUTION;
	// Never delay more than a whole second. This indicates a failure of
	// some part of the system (eg: m_delay has dropped below 0) so warn
	// about it.
				if (wait > 1000) 
				{
	fprintf(stderr, "wait=%ld capped to 1000  m_delay=%ld delta=%ld "
		"m_ticks=%ld t=%ld\n", wait, m_delay, delta, m_ticks, t);
					wait = 1000;
				}
				if (wait < 0) wait = 0;
				poll_events();
				SDL_Delay(wait);
				m_delay -= wait;
			}	
			m_ticks = t;
			m_timers = 0;	
		}
	}

/* All the commented-out code is stuff to turn on the CPU debugging modes 
 * under particular conditions. It was used to track down emulation problems
 * without generating instruction traces a mile long. */

//	unsigned char *watch = mem_base + 0x1503A;
//	unsigned char *watch = mem_base + 0xF039;
//	unsigned char *drvr_cs = m_mem_base + 0x3AF50;
//	unsigned char *drvr_ds = m_mem_base + 0x467D0;
//	static int owatch = -1;
	static int go = 0;
//	static int watch_seg = 0;
//	static int watch_off = 0;
//	static int watch_last = 0;
/*
	if (drvr_cs[0] == 0x56 && drvr_cs[1] == 0x9C)
	{
		if (R_CS == 0x3AF5 && R_IP == 0x0C)
		{
			fprintf(stderr, "Driver call with DS=%04x "
					"MONO8XHT_MODE=%04x\n", R_AX,
					fetch16(R_AX, 0x0F20));
			go = 1;
			watch_seg = R_AX;
			watch_off = 0x0F20;
			watch_last = fetch16(watch_seg, watch_off);
		}
		if (go == 1 &&
		    fetch16(watch_seg, watch_off) != watch_last)
		{
			watch_last = fetch16(watch_seg, watch_off);
			fprintf(stderr, "MONO8XHT_MODE smashed!\n");
			FILE *fp = fopen("apricore", "wb");
			fwrite(m_mem_base, 1, 1048576, fp);
			fclose(fp);
			settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
		}
	
		if (R_CS == 0x3AF5 && R_IP == 0xDD)
		{
			fprintf(stderr, "Driver retn with DS=%04x "
					"MONO8XHT_MODE=%04x\n", R_DS,
					fetch16(R_DS, 0x0F20));
			go = 0;
		}
	}
*/
/* [30-Aug-2008] Trace the game SNAKE. Compiled BASIC, ick */
/*
	if (R_CS == 0x26a2 && R_IP == 0x02f4)
	{
		fprintf(stderr, "Test bit:\n");
		fprintf(stderr,	"\tArray size=%02x\n",   fetch16(R_DS, 8 + R_BP));
		fprintf(stderr, "\tArray base=%04x\n", fetch16(R_DS, 6 + R_BP));
		fprintf(stderr,	"\tBit       =%02x\n",   fetch16(R_DS, 10 + R_BP));
		fprintf(stderr,	"\tByte      =%02x\n", fetch(R_DS, R_SI));
		fprintf(stderr,	"\tBitno     =%02x\n",   R_CX & 0xFF);
	}
	if (R_CS == 0x26a2 && R_IP == 0x02fd)
	{
		fprintf(stderr,	"\tResult    =%04x\n",   R_AX);
	}
	if (R_CS == 0x26d6 && R_IP == 0x0940)
	{
		settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
	}
	if (R_CS == 0x26d6 && R_IP == 0x094e)
	{
		settings.debug &= ~(DEBUG_DECODE_F | DEBUG_TRACE_F);
	}
	if (R_CS == 0x21cd && R_IP == 0x0d57)
	{
		settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
	}
	if (R_CS == 0x1880 && R_IP == 0x621)
	{
		fprintf(stderr, "INT 21 AX=%04x DS:DX=%04x:%0x ",
				R_AX, R_DS, R_DX);
		if ((R_AX & 0xFF00) == 0x0f00)
		{
			fprintf(stderr, "f_open(");
			for (int n = 1; n <= 11; n++)
			{
				fprintf(stderr, "%c", fetch(R_DS, R_DX + n));	
			}
			fprintf(stderr, ")\n");
		}
		else if ((R_AX & 0xFF00) == 0x1000)
		{
			fprintf(stderr, "f_close(");
			for (int n = 1; n <= 11; n++)
			{
				fprintf(stderr, "%c", fetch(R_DS, R_DX + n));	
			}
			fprintf(stderr, ")\n");
		}
		else fprintf(stderr, "\n");
	}
*/
/* [27-Aug-2008] Trace FP startup */
#if 0
	static int ready = 0, watch = -1;
	if (R_CS == 0xf8f2 && R_IP == 0x024b)
	{
		ready = 1;
		watch = fetch(0xD000, 0);
		fprintf(stderr,"ready=1 peek(0xd000)=%02x\n", watch);
//		settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
	}
	if (ready == 1 && fetch(0xD000, 0) != watch)
	{
		fprintf(stderr,"alert! peek(0xd000)=%02x\n", fetch(0xd000,0));
		ready = 0;
		int j = fetch(0xd000, 0);
//		settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
	}
	if (R_CS == 0xf8f2 && R_IP == 0x02b5)
	{
		fprintf(stderr, "fail!\n");
		settings.debug &= ~(DEBUG_DECODE_F | DEBUG_TRACE_F);
	}
#endif

/* [27-Aug-2008] Monitor what the mouse driver's doing, in an attempt to 
 * discover why, however you move the mouse, the pointer just goes down.
 *
 * Conclusion: The driver's bugged and QDAE is not doing anything wrong.

	if (R_CS == 0x0E00 && R_IP == 0x5D90)
	{
		fprintf(stderr, "Packet received: ");
		int n;
		for (n = 0; n < 5; n++)
		{
			fprintf(stderr, "%02x ", fetch(R_DS, R_BX + n));
		}
		fprintf(stderr, "\n");
		settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
	}
	if (R_CS == 0x0E00 && R_IP == 0x5DA8)
	{
		settings.debug &= ~(DEBUG_DECODE_F | DEBUG_TRACE_F);
	}
	if (R_CS == 0x0E00 && R_IP == 0x5DAA)
	{
		settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
	}
	if (R_CS == 0x0E00 && R_IP == 0x5DB8)
	{
		settings.debug &= ~(DEBUG_DECODE_F | DEBUG_TRACE_F);
	}

	if (R_CS == 0x1CB1 && R_IP == 0x08F3)
	{
		fprintf(stderr, "Enter INT F5 handler, AX=%04x\n",
				R_AX);
		if (R_AX & 0x800) fprintf(stderr, "yyy ");
		else		  fprintf(stderr, "??? ");
		if (R_AX & 0x400) fprintf(stderr, "Sw2 ");
		else		  fprintf(stderr, "    ");
		if (R_AX & 0x200) fprintf(stderr, "Sw1 ");
		else		  fprintf(stderr, "    ");

		if (R_AX & 0x100) fprintf(stderr, "Y=%d\n", (R_AX & 0xFF));
		else		  fprintf(stderr, "X=%d\n", (R_AX & 0xFF));
//		settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
	}
//	if (R_CS == 0x1cb1 && R_IP == 0x0bb1)
//	{
//		fprintf(stderr, "Leave INT F5 handler\n");
//		settings.debug &= ~(DEBUG_DECODE_F | DEBUG_TRACE_F);
//	} 
*/
/* [20-Aug-2008] Debug CCP/M-86 startup  

	if (R_CS == 0x0e5b && R_IP == 0x03E2)
	{
		fprintf(stderr, "Enter dispatcher; RLR=%04x CS:IP=%04x:%04x\n",
				fetch16(0x14e08),
				fetch16(R_SS, R_SP+2),
				fetch16(R_SS, R_SP));
	}
	if (R_CS == 0x0e5b && R_IP == 0x045F)
	{
		fprintf(stderr, "Leave dispatcher; RLR=%04x CS:IP=%04x:%04x\n",
				fetch16(0x14e08),
				fetch16(R_SS, R_SP+2),
				fetch16(R_SS, R_SP));
	}
*/

	//if (R_CS == 0x1165 && R_IP == 0x095f)
	//if (R_CS == 0x0d00 && R_IP == 0x1cfe)
	//{
	//	settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
	//}
/* [20-Oct-2007] Debug trace of mouse movement 

	if (R_CS == 0xE00 && R_IP >= 0x5D5D && R_IP < 0x5DBC)
	{
		settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
	}
	else
	{
		settings.debug &= ~(DEBUG_DECODE_F | DEBUG_TRACE_F);
		fprintf(stderr, "Leave dispatcher:
	}*/

/*
	if (R_CS == 0x0e5b)
	{
		settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
	} */
/*	else	
	{
		settings.debug &= ~(DEBUG_DECODE_F | DEBUG_TRACE_F);
	} 
*/
	//if (R_CS == 0x1165 && R_IP == 0x095f)
	//if (R_CS == 0x0d00 && R_IP == 0x1cfe)
	//{
	//	settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
	//}
/* [20-Oct-2007] Debug trace of mouse movement 

	if (R_CS == 0xE00 && R_IP >= 0x5D5D && R_IP < 0x5DBC)
	{
		settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
	}
	else
	{
		settings.debug &= ~(DEBUG_DECODE_F | DEBUG_TRACE_F);
	}
*/
	/* 

	if (R_CS == 0xecf0 && R_IP == 0x06DC)
	{
		FILE *fp = fopen("debug_on", "r");
		if (fp)
		{
			settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
			fclose(fp);
		}
	}
*/
		
/*	if (R_CS == 0x0fd10 && R_IP == 0x1D1F)
	{
		printf("System font initialised\n");
		fflush(stdout);
	}
	if (R_CS == 0x0c80 && R_IP == 0x113)
	{
		settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
		go = 1;
	}
	if (R_CS == 0x0c80 && R_IP == 0x11A)
	{
		settings.debug &= ~(DEBUG_DECODE_F | DEBUG_TRACE_F);
		go = 1;
	}
*/
//	if (R_CS == 0x0f00 && R_IP == 0x893)
//	{
//		printf("Break me here\n");
//		debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
//	}

/*	if (watch[0] != owatch)
	{
		fprintf(stderr, "CS=%04x IP=%04x 0x1503A changes from "
				"%02x to %02x\n", R_CS, R_IP, 
				owatch, watch[0]);
		if (go && watch[0] == 0)
		{
			exit(1);
		}
		owatch = watch[0];
	}
*/

	if (go == 2) // R_IP == 0x31F8 && R_CS == 0xE00) 
	{
		FILE *fp;
		fp = fopen("apricore", "wb");
		fwrite(m_mem_base, 1, 1048576, fp);
		fclose(fp);
		fprintf(stderr, "Apricot crash - core dumped\n");
		exit(1);
//		debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;

	}

}

UiEvent PC_ENV::dispatch_event(SDL_Event *e)
{
	if (m_keyboard->keyboard_event(e) == UIE_QUIT) exit(0);

	return UIE_CONTINUE;
}

void PC_ENV::poll_events()
{
	SDL_Event e;
	if (SDL_PollEvent(&e))
	{
		dispatch_event(&e);
	}
}

bool PC_ENV::illegal_op()
{
	UiEvent e = gl_ui->illegal_op();

	if (e == UIE_QUIT) return true;

	return false;
}


bool PC_ENV::halted()
{
	UiEvent e = gl_ui->halted();

	if (e == UIE_QUIT) return true;

	return false;
}


void	sys_fatal(int error, char *fmt, ...)
{
  va_list	p;
	
  va_start(p, fmt);
  fprintf(stderr, "Fatal error: ");
  if (error != 0)
      {
	fprintf(stderr, "<%d>",error);
	fprintf(stderr,"%s",STRERROR(error));
      }

  vfprintf(stderr, fmt, p);
  va_end(p);
  fprintf(stderr, "\nExiting...\n");
  exit(1);
}

void	sys_warn(char *fmt, ...)
{
  va_list	p;
	
  va_start(p, fmt);
  fprintf(stderr, "Warning: ");
  vfprintf(stderr, fmt, p);
  va_end(p);
}


/* Note that this is a shallow copy! */
void PC_ENV::copy_to(PC_ENV *other)
{
	other->Gn_regs = Gn_regs;
	other->Sp_regs = Sp_regs;
	other->Sg_regs = Sg_regs;
	other->settings = settings;
	other->sysmode = sysmode;
	other->m_intNo = m_intNo;
	other->m_intDev = m_intDev;
	other->m_mem_base = m_mem_base;
	other->dbg_stream = dbg_stream;
#ifdef DEBUG
	other->check = check;
	other->saved_ip = saved_ip;
	other->saved_cs = saved_cs;
	other->enc_pos = enc_pos;
	other->enc_str_pos = enc_str_pos;
	memcpy(other->decode_buf, decode_buf, sizeof(decode_buf));
	memcpy(other->decoded_buf, decoded_buf, sizeof(decoded_buf));
#endif
}


void PC_ENV::load_rom(const char *filename, int address, int len,
		      const char *altname)
{
	FILE *fp;
	std::string error;

	fp = openQdaeFile(FT_ROM, filename, "rb");
	if (altname != NULL && !fp)
	{
		filename = altname;
		fp = openQdaeFile(FT_ROM, filename, "rb");
	}
	if (!fp)
	{
		error = "Could not open ";
		error += filename;
		gl_ui->report(error.c_str(), "  Exit  ");
		exit(1);
	}

	if (fread(m_mem_base + address, 1, len, fp) < len)
	{
		error = "Warning: ";
		error += filename;
		error += " is shorter than expected.\n";
		fprintf(stderr, error.c_str());
	}
	fclose(fp);
}


//////////////////////////////////////////////////////////////////////////////
//
// Bring up settings pages
//
UiEvent  PC_ENV::settingsMenu(UiDrawer *d, int x, int y)
{
	vector<int> map;
	UiMenu uim(d);
	UiEvent uie;
	SDLKey k;
	int entries, n;
	string caption;

	for (n = entries = 0; n < m_devCount; n++)
	{
		if (m_devices[n]->hasSettings(&k, caption))
		{
			map.push_back(n);
			uim.add(new UiCommand(k, caption,d, UIG_SUBMENU));
		}
	}
	uim.add(new UiCommand(SDLK_g,      "  General settings  ", d))
	   .add(new UiSeparator(d))
	   .add(new UiCommand(SDLK_s,      "  Save settings  ", d))
	   .add(new UiCommand(SDLK_ESCAPE, "  Exit", d));

	uim.setX(x);
	uim.setY(y);
	do
	{
		uie = uim.eventLoop();
		if (uie == UIE_QUIT || uie == UIE_CANCEL) return uie;

		n = uim.getSelected();
		if (n >= (int)map.size()) 
		{
			if (uim.getKey(n) == SDLK_ESCAPE) break;
			if (uim.getKey(n) == SDLK_g)
			{
				uie = settings.settings(d);
			}	
			if (uim.getKey(n) == SDLK_s) 
			{
				storeHardware();
				if (saveHardware()) uie = gl_ui->report("Settings were saved", "  OK");
				else		    uie = gl_ui->report("Failed to save settings","  Cancel");
 
				if (uie == UIE_QUIT) return uie;
			}
		}
		else
		{
			uie = m_devices[map[n]]->settings(d);
			if (uie == UIE_QUIT) return uie;
		}			
	} while(1);	
	return UIE_OK;
}

void PC_ENV::addDevice(Device *d)
{
	int n;

	if (m_devCount >= m_devMax)
	{
		Device **t = new Device *[2 * m_devMax];
	
		if (!t) return;
		for (n = 0; n < m_devCount; n++) t[n] = m_devices[n];
		delete m_devices;
		m_devices = t;
		m_devMax *= 2;
	}
	m_devices[m_devCount++] = d;
}

//
// Access to XML settings file
//
bool PC_ENV::loadHardware()
{
        string filename;
        xmlDocPtr doc;
        xmlNsPtr ns;
        xmlNodePtr cur;

        filename = findQdaeFile(FT_NORMAL, m_hwfile, "r");

//
// Do a lot of sanity checks on the hardware XML.
//
        doc = xmlParseFile(filename.c_str());
        if (doc == NULL)	// Can't open it
        {
                fprintf(stderr, "Could not open %s\n", filename.c_str());
                return 0;
        }
        cur = xmlDocGetRootElement(doc);
        if (cur == NULL)	// No root element
        {
                fprintf(stderr, "%s: File is empty\n", filename.c_str());
                xmlFreeDoc(doc);
                return 0;
        }
        ns = xmlSearchNsByHref(doc, cur, (xmlChar *)m_xmlns);
        if (ns == NULL)		// No namespace
        {
                fprintf(stderr, "%s: No QDAE namespace found\n", filename.c_str());
                xmlFreeDoc(doc);
                return 0;
        }
        if (strcmp((char *)cur->name, "qdae")) // Wrong root
        {
                fprintf(stderr, "%s: root node is not of type 'qdae'.", filename.c_str());
                xmlFreeDoc(doc);
                return 0;
        }
	// Newly-loaded DOM is okay.
	if (m_settings) xmlFreeDoc(m_settings);
	m_settings = doc; 
	return true;
}



bool PC_ENV::parseHardware()
{
	xmlNodePtr root = xmlDocGetRootElement(m_settings);
	xmlNsPtr   ns   = xmlSearchNsByHref(m_settings, root, (xmlChar *)m_xmlns);
	bool okay = true;

//	XLOG("PC_ENV::parseHardware()");
        for (int n = 0; n < m_devCount; n++) 
	{
		okay = okay && m_devices[n]->parse(m_settings, ns, root);
	}
	return okay;
}


bool PC_ENV::storeHardware()
{
        xmlNodePtr root = xmlDocGetRootElement(m_settings);
        xmlNsPtr   ns   = xmlSearchNsByHref(m_settings, root, (xmlChar *)m_xmlns);
	bool okay = true;

        for (int n = 0; n < m_devCount; n++) 
        {
                okay = okay && m_devices[n]->store(m_settings, ns, root);
        }
	return okay;
}


bool PC_ENV::saveHardware()
{
	string filename;

	if (!m_settings) return false;

        filename = findQdaeFile(FT_NORMAL,m_hwfile,"w"); 

        if (xmlSaveFile(filename.c_str(), m_settings) < 0) return false;

	if (!settings.saveConfig()) return false;
        return true;
}


