/* 
 * Dos/PC Emulator
 * Copyright (C) 1991 Jim Hudgens
 * Copyright 2005,2007,2008 John Elliott 
 * 
 * The file is part of GDE.
 * 
 * GDE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * GDE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GDE; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 *
 */

class Keyboard;
class Video;
class Memsize;
class Speed;

class PC_ENV
{
private:
	Device **m_devices;
	int m_devCount, m_devMax;
	xmlDocPtr m_settings;
	const char *m_hwfile;
	const char *m_xmlns;
	long	m_timers, m_ticks, m_delay;
public:
	PC_ENV(const char *hwfile, const char *xmlns);
	virtual ~PC_ENV();

	virtual void initialize();
	virtual void reboot();
	virtual void clock_tick();

	bool loadHardware();	// Load the hardware description XML file.
	virtual bool parseHardware();	// Allow the modules to parse the XML DOM.
	bool storeHardware();	// Ask the modules to save their state to the DOM
	bool saveHardware();	// Save the hardware XML file.

/* Display the settings menu */
	UiEvent settingsMenu(UiDrawer *d, int x, int y);

/* Floppy access */
	virtual void floppy_eject(int drive) = 0;
	virtual UiEvent floppy_insert(int drive, std::string filename,
			const char *type) = 0;

/* I/O ports differ across models */
	virtual void sys_out8(u_int16 port, u_int8 value) = 0;
	virtual void sys_out16(u_int16 port, u_int16 value) = 0;
	virtual u_int16 sys_in16(u_int16 port) = 0;
	virtual u_int8 sys_in8(u_int16 port) = 0;
	virtual void sys_lock(void) = 0;

/* Return a shallow copy of this object */
	virtual PC_ENV *clone() = 0;

	virtual void poll_events(void);
	virtual UiEvent dispatch_event(SDL_Event *e);

/* Deprecated. Use only for divide overflow and NMI. */
	void	i86_intr_raise(u_int8 intrnum);
/* All other interrupts should do this instead: */
	void	i86_intr_raise(IntAck *device);
	void	i86_intr_lower(IntAck *device);
/* Set the processor going */
	void	i86_exec(void);

protected:
	void addDevice(Device *d);
	void debug_init();
	void copy_to(PC_ENV *m);
	void load_rom(const char *filename, int address, int len,
			const char *altname = NULL);

	void i86_intr_handle();
public:
   /*	The registers!!		*/
   struct i386_general_regs  Gn_regs;
   struct i386_special_regs  Sp_regs;
   struct i386_segment_regs  Sg_regs;
   /* our flags structrure.  This contains information on
           REPE prefix           2 bits  repe,repne
           SEGMENT overrides     5 bits  normal,DS,SS,CS,ES
           Delayed flag set      3 bits  (zero, signed, parity)
	   reserved              6 bits
           interrupt #           8 bits  instruction raised interrupt
	   BIOS video segregs    4 bits  
	   Interrupt Pending     1 bits  
           Extern interrupt      1 bits
	   Halted                1 bits
    */
   int m_cputicks;	   
   Speed     *m_speed;
   Keyboard *m_keyboard;
   Video    *m_video;
   Memsize  *m_memsize;
   GeneralSettings settings;   
   long   sysmode;
   int	  m_intNo;
   IntAck *m_intDev;

   inline void store(u_int16 seg, u_int16 off, u_int8 value)
   {
	store(  ((((u_int32)seg) << 4) + off) & 0xFFFFF, value); 
   }

   inline u_int8 fetch(u_int16 seg, u_int16 off)
   {
	return fetch(  ((((u_int32)seg) << 4) + off) & 0xFFFFF);
   }

   virtual void store(u_int32 address, u_int8 value) = 0;
   virtual u_int8 fetch(u_int32 address) = 0;
   
   inline u_int16 fetch16(u_int16 seg, u_int16 off)
   {
/* Note 2 calls to fetch, not one to fetch16, so that wrap at off=0xFFFF
 * works */
	return fetch(seg, off) | (((u_int16)fetch(seg, off+1)) << 8);
   }

   inline void store16(u_int16 seg, u_int16 off, u_int16 v)
   {
	store(seg, off,   (v & 0xFF));
	store(seg, off+1, (v >> 8));
   }
   
   inline void store16(u_int32 address, u_int16 v)
   {
	store(address,   (v & 0xFF));
	store(address+1, (v >> 8));
   }

   inline u_int16 fetch16(u_int32 address)
   {
	return fetch(address) | (((u_int16)fetch(address+1)) << 8);
   }
   
   inline u_int32 fetch32(u_int32 address)
   {
	return fetch16(address) | (((u_int32)fetch16(address+2)) << 16);
   }
  
   /* Get a segment:offset pair as an absolute offset */ 
   inline u_int32 fetch20(u_int32 address)
   {
	return fetch16(address) + (((u_int32)fetch16(address+2)) << 4);
   }


   /* Illegal opcode. Return true to abort, false to continue (possibly with
    * fixed-up registers) */
   virtual bool illegal_op();

   /* System halted. Return true to abort, false to continue */
   virtual bool halted();

   /* system memory information */
   u_int8 *	m_mem_base;
	
#ifdef DEBUG
   int			check;
   u_int16		saved_ip;
   u_int16		saved_cs;
   int                  enc_pos;
   int                  enc_str_pos;
   char                 decode_buf[32];      /* encoded byte stream  */
   char                 decoded_buf[256];   /* disassembled strings */
#endif
   /* needs to be defined always. */
   FILE                 *dbg_stream;
};


