/************************************************************************* 
 
    Quick And Dirty Apricot Emulator
    Copyright (C) 2005, 2007  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "qdae.hxx"
#include "qdae_ui.hxx"
#include "qdae_files.hxx"
#include "sdl_video.hxx"
#include "xi_video.hxx"
#include "SdlContext.hxx"
#include "SdlTerm.hxx"
#include "sdl_keyboard.hxx"
#include "wd2797.hxx"

#include "UiContainer.hxx"
#include "UiLabel.hxx"
#include "UiSeparator.hxx"
#include "UiSetting.hxx"
#include "UiCommand.hxx"
#include "UiMenu.hxx"
#include "UiColourEntry.hxx"


#define SCREEN_W 800
#define SCREEN_H 400
#define SCREEN_DEPTH 32
#define SCREEN_FLAGS 0
#define SCREEN_EXTRA 200
#define FPS_DELAY 100	/* Giving 10fps */
#define VRAMBASE 0x800
#define VRAMLEN  0x1000
#define PALETTE 0xE0000
#define VPTRBASE 0x1E00
#define VPTRLEN  0x1000

//#define VIDEO_LOG(x)  printf x ; fflush(stdout);
#define VIDEO_LOG(x)


/* In text mode, video RAM is an 80x25 array of 16-bit words with bits 15-11
 * being attributes and bits 10-0 being character. Character bitmaps are 
 * 10x16.
 * 	Attribute bits: Bit 12 = strikethrough
 * 			Bit 13 = underline
 * 			Bit 14 = Bright
 * 			Bit 15 = Reverse
 *
 * In character bitmaps, bit 14 of character row is set to draw strikethrough
 * there, and bit 15 set to draw underline there.
 *
 * In graphics mode it's a 50x25 array but attributes are ignored. 
 * Character bitmaps are 16x16.
 *
 * The bitmap pointers live at F000:0 to F000:FFF
 */

XI_Video::XI_Video(PC_ENV *m) : Video(m, "xi", SCREEN_W, SCREEN_H + SCREEN_EXTRA)
{
	memset(shadowchar, 0xFF, sizeof(shadowchar));
	extra_height = 16;
	vscale = 1;
	scr_rows = 400;
	enable(false);	
	m_uscreen = NULL;
}

XI_Video::~XI_Video()
{
	if (m_uscreen) delete m_uscreen;
}

u_int8 XI_Video::read(u_int16 port)
{
	switch(port)
	{
		case 0:	return m_crtc.m_cur;
		case 1: return m_crtc.get();
	}
	return 0xFF;
}

void XI_Video::write(u_int16 port, u_int8 value)
{
	switch(port)
	{
		case 0:
			m_crtc.m_cur = value;
			break;
		case 1:
			VIDEO_LOG(("Video: CRTC[%d] = %d\n", 
					m_crtc.m_cur, value));
			m_crtc.set(value);

/* CRTC regs 12/13: Video start address */
			switch(m_crtc.m_cur)
			{
				case 10: m_curtop = value & 0x1F;
					 m_curmode = value & 0xC0;
					 switch(m_curmode)
					 {
						case 0x80: 
							m_curcount = 0;
							m_curmax = 32;
							break;
						case 0xC0:
							m_curcount = 0;
							m_curmax = 64;
							break;
					 }
					 break;
				case 11: m_curbot = value & 0x1F;
					 break;
				case 12: m_base &= 0x00FF;
					 m_base |= ((u_int16)(value & 0x3F)) << 8;
					 break;
				case 13: m_base &= 0xFF00;
					 m_base |= value;
					 break;
				case 14: m_curaddr &= 0x00FF;
					 m_curaddr |= ((u_int16)value) << 8;
					 break;
				case 15: m_curaddr &= 0xFF00;
					 m_curaddr |= value;
					 break;
			}
			break;
#if 0
		case 13: /* Set Stop LED */
			gl_stop = value;
			gl_ledsdirty = 1;
			break;

		case 11: /* Set Caps Lock LED */
			gl_caps = value;
			gl_ledsdirty = 1;
			break;

		case 9: /* Set horizontal resolution */
			v_mode = value;
			gl_dirty = 1;
			break;

		case 7: /* Set vertical resolution */
			gl_dirty = 1;
			if (value == 0) 
			{
				SDL_Rect rc;
	
				scr_rows = 200;
				rc.x = 0;
				rc.y = 200 * vscale;
				rc.w = SCREEN_W;
				rc.h = (SCREEN_H - 200) * vscale;
				SDL_FillRect(vidSurface, &rc, m_bg);
			}
			else	scr_rows = 256;
			break;
#endif
	}
}

void XI_Video::paint_gfx(int y, int x, Uint16 ch)
{
	int y1, x1;
	unsigned char mask;
	unsigned char *font;
	unsigned char *framebuf;
	unsigned offset, xoffset;
	unsigned char *bitmap;
	unsigned bpp;
	int col;
	u_int8 cur;
	Uint32 fg, bg;

	bpp = vidSurface->format->BytesPerPixel;
       	framebuf = (unsigned char *)(vidSurface->pixels);
	font = sysram + ((ch & 0x7FF) << 5);

	fg = m_fg;
	bg = m_bg;

	offset = (y * 16 * vidSurface->pitch) + (x * 16 * bpp);
	for (y1 = 0; y1 < 16; y1++)
	{
		xoffset = offset + (y1 * vidSurface->pitch);

		mask = 0x01;
		cur = font[2*y1];
		for (x1 = 0; x1 < 16; x1++)
		{
			Uint32 *ptr = (Uint32 *)(framebuf + xoffset);
			*ptr = (cur & mask) ? fg : bg;
			mask = mask << 1;
			if (mask == 0) 
			{
				mask = 1;	
				cur = font[2*y1 + 1];
			}
			xoffset += bpp;
		}
	}
}

void XI_Video::paint_text(int y, int x, Uint16 ch, bool cursor)
{
	int y1, x1;
	unsigned char mask;
	unsigned char *font;
	unsigned char *framebuf;
	unsigned offset, xoffset;
	unsigned char *bitmap;
	unsigned bpp;
	int col;
	Uint32 fg, bg, tmp;
	bool ul = false, strike = false;

	bpp = vidSurface->format->BytesPerPixel;
       	framebuf = (unsigned char *)(vidSurface->pixels);
	font = sysram + ((ch & 0x7FF) << 5);

	fg = m_fg;
	bg = m_bg;

	strike = (ch & 0x1000) != 0;
	ul     = (ch & 0x2000) != 0;

	if (ch & 0x4000)	// Bright
	{
		fg = m_bright;
	}
	if (ch & 0x8000)	// Reverse video
	{
		tmp = bg; bg = fg; fg = tmp;	
	}
	if (cursor)
	{
		switch(m_curmode)
		{
			case 0x40: cursor = false; break;
			case 0x80: cursor = (m_curcount < 16); break;
			case 0xC0: cursor = (m_curcount < 32); break;
		}
	}

	offset = (y * 16 * vidSurface->pitch) + (x * 10 * bpp);
	for (y1 = 0; y1 < 16; y1++)
	{
		xoffset = offset + (y1 * vidSurface->pitch);

		mask = 0x01;
/* Check for underline and strikethrough */
		u_int8 line = 0;
		u_int8 flags = font[2*y1 + 1] & 0xC0;

		if (ul     && (flags & 0x80)) line = 0xFF;
		if (strike && (flags & 0x40)) line = 0xFF;
		u_int8 cur = font[2*y1] | line; 
		if (cursor && y1 >= m_curtop && y1 <= m_curbot)
			cur = ~cur;		

		for (x1 = 0; x1 < 10; x1++)
		{
			Uint32 *ptr = (Uint32 *)(framebuf + xoffset);
			*ptr = (cur & mask) ? fg : bg;
			mask = mask << 1;
			if (mask == 0) 
			{
				mask = 1;	
				cur = font[2*y1 + 1] | line;
				if (cursor && y1 >= m_curtop && y1 <= m_curbot)
					cur = ~cur;		
			}
			xoffset += bpp;
		}
	}
}	


int XI_Video::paint()
{
	int x, y;
	int any;
	long offset, shadow_offset;
	SDL_Event e;
	Uint8 *ch, *cur;

	gl_dirty = 1;

	for (;;)
	{
		SDL_Delay(FPS_DELAY);

		// If emulator UI is using the screen, don't paint.
		if (gl_ui->usingScreen()) continue;
/* Paint MicroScreen */
		if (gl_dirty && m_uscreen)
		{
			m_uscreen->repaint(0, SCREEN_H, SCREEN_W,
					SCREEN_EXTRA - 16);
		}
		if (m_uscreen)
		{
			m_uscreen->updateTime();
		}
		if (!isEnabled())
		{
			SDL_Rect rc;
	
			rc.x = 0;
			rc.y = 0;
			rc.w = SCREEN_W;
			rc.h = SCREEN_H;
			SDL_FillRect(vidSurface, &rc, m_bg);
			SDL_UpdateRect(vidSurface, 0, 0, 0, 0);
			continue;	
		}
		if (m_curmode & 0x80)
		{
			++m_curcount;
			m_curcount %= m_curmax;
		}

/* XXX Only repaint a cell if it's actually changed! */
		SDL_LockSurface(vidSurface);
		for (y = 0; y < 25; y++)
		{
			if (m_text_mode)
			{
				offset = m_curaddr % 2048;
				cur = sysram + 0xF0000 + (2 * offset);
/* Where the cursor should be drawn: XXX implement this! */
				for (x = 0; x < 80; x++)
				{
					offset = (m_base + 80 * y + x) % 2048;
					ch = sysram + 0xF0000 + 2 * offset;
					u_int16 cv = (ch[1] << 8) | ch[0];
					if (shadowchar[y][x] != cv || gl_dirty)
					{
						shadowchar[y][x] = cv;
					}
					paint_text(y, x, cv, cur == ch);
				}
			}
			else
			{
				for (x = 0; x < 50; x++)
				{
					offset = (m_base + 50 * y + x) % 1250;
					ch = sysram + 0xF0000 + 2 * offset; 
					u_int16 cv = (ch[1] << 8) | ch[0];
					if (shadowchar[y][x] != cv || gl_dirty)
					{
						shadowchar[y][x] = cv;
					}
					paint_gfx(y, x, cv);
				}	
			}
		}
		SDL_UpdateRect(vidSurface, 0, 0, 0, 0);
		SDL_UnlockSurface(vidSurface);

		if (gl_dirty || gl_ledsdirty)
		{
			term->setForeground(0, 0xFF, 0);
			term->moveCursor(0, 0);
			(*term) << (gl_caps ? "CAPS " : "     ");
			term->setForeground(0xFF, 0xFF, 0);
			(*term) << (gl_stop ? "STOP " : "     ");
			term->setForeground(0xFF, 0x00, 0);
			(*term) << (floppy_indicator[0] ? "A " : "  ");
			(*term) << (floppy_indicator[1] ? "B " : "  ");
			gl_ledsdirty = 0;
		}

		gl_dirty = 0;
	}
}

void XI_Video::init()
{
	int n;

	Video::init();
	m_uscreen = new Microscreen(vidSurface);
	m_text_mode = true;
	m_base = 0;

	sysram = host->m_mem_base;
	m_clrbg.r = 0x00; m_clrbg.g = 0x00; m_clrbg.b = 0x00;
	m_clrfg.r = 0x00; m_clrfg.g = 0xc0; m_clrfg.b = 0x00;
	m_clrbr.r = 0x00; m_clrbr.g = 0xff; m_clrbr.b = 0x00;
	m_bg = SDL_MapRGB(vidSurface->format, m_clrbg.r, m_clrbg.g, m_clrbg.b);
	m_fg = SDL_MapRGB(vidSurface->format, m_clrfg.r, m_clrfg.g, m_clrfg.b);
	m_bright = SDL_MapRGB(vidSurface->format, m_clrbr.r, m_clrbr.g, m_clrbr.b);

}

void XI_Video::setTextMode(bool b)
{
	if (b == m_text_mode) return;

	m_text_mode = b;
	gl_dirty = true;
}


bool XI_Video::parseNode(xmlDoc *doc, xmlNs *ns, xmlNode *node)
{
	if (!parseColour(doc, ns, node, "bright", m_clrbr, m_clrbg)) 
		return false;
	if (!parseColour(doc, ns, node, "normal", m_clrfg, m_clrbg))
		return false;
	m_bg = SDL_MapRGB(vidSurface->format, m_clrbg.r, m_clrbg.g, m_clrbg.b);
	m_fg = SDL_MapRGB(vidSurface->format, m_clrfg.r, m_clrfg.g, m_clrfg.b);
	m_bright = SDL_MapRGB(vidSurface->format, m_clrbr.r, m_clrbr.g, m_clrbr.b);
	return true;
}


bool XI_Video::storeNode(xmlDoc *doc, xmlNs *ns, xmlNode *node)
{
	if (!storeColour(doc, ns, node, "bright", m_clrbr, m_clrbg)) 
		return false;
	return storeColour(doc, ns, node, "normal", m_clrfg, m_clrbg); 
}


bool XI_Video::hasSettings(SDLKey *key, string &caption)
{
	*key = SDLK_v;
	caption = "  Video  ";
	return true;
}


UiEvent XI_Video::customClrs(UiDrawer *d)
{	
	UiEvent uie;
	int sel = -1;
 
        while (1)
        {
                UiMenu uim(d);
                uim.add(new UiLabel("  Mono screen colours  ", d));
                uim.add(new UiSeparator(d));
		uim.add(new UiColourEntry(SDLK_f, "  Foreground:  __  ", m_clrfg, d));
		uim.add(new UiColourEntry(SDLK_b, "  Background:  __  ", m_clrbg, d));
		uim.add(new UiColourEntry(SDLK_r, "  Bright: __  ", m_clrbr, d));
                uim.add(new UiSeparator(d));
		uim.add(new UiCommand(SDLK_ESCAPE, "  EXIT  ", d));
                d->centreContainer(uim);
                uim.setSelected(sel);
		uie = uim.eventLoop();
		
		if (uie == UIE_QUIT) return uie;
                sel = uim.getSelected(); 
		switch(uim.getKey(sel))
		{
			case SDLK_ESCAPE:
			return UIE_OK;
			default:		break;
		}
	} 
	return UIE_CONTINUE;
}


UiEvent XI_Video::settings(UiDrawer *d)
{
	int sel = -1;
	UiEvent e;
	bool green, white, amber, blkbg;

        while (1)
        {
		blkbg = (m_clrbg.r == 0 && m_clrbg.g == 0 && m_clrbg.b == 0);
		green = (m_clrfg.r == 0x00 && m_clrbr.r == 0x00 &&
			 m_clrfg.g == 0xc0 && m_clrbr.g == 0xff &&
			 m_clrfg.b == 0x00 && m_clrbr.b == 0x00 && blkbg);
		white = (m_clrfg.r == 0xc0 && m_clrbr.r == 0xff &&
			 m_clrfg.g == 0xc0 && m_clrbr.g == 0xff &&
			 m_clrfg.b == 0xc0 && m_clrbr.b == 0xff && blkbg);
		amber = (m_clrfg.r == 0xc0 && m_clrbr.r == 0xff &&
			 m_clrfg.g == 0x80 && m_clrbr.g == 0xa0 &&
			 m_clrfg.b == 0x00 && m_clrbr.b == 0x00 && blkbg);

                UiMenu uim(d);
                bool bEnabled;

                bEnabled = isEnabled();
                uim.add(new UiLabel("  Video settings  ", d));
                uim.add(new UiSeparator(d));
                uim.add(new UiSetting(SDLK_g, green, "  Green screen  ", d));
                uim.add(new UiSetting(SDLK_w, white, "  White screen  ", d));
                uim.add(new UiSetting(SDLK_a, amber, "  Amber screen  ", d));
                uim.add(new UiSetting(SDLK_c, !(green||white||amber), "  Custom colours...  ", d));
	       	uim.add(new UiSeparator(d));
	        uim.add(new UiCommand(SDLK_ESCAPE, "  EXIT  ", d));
		uim.setSelected(sel);
		d->centreContainer(uim);	
		e = uim.eventLoop();
                if (e == UIE_QUIT) return e;
                sel = uim.getSelected();
                switch(uim.getKey(sel))
		{
			case SDLK_c:	e = customClrs(d);
					if (e == UIE_QUIT) return e;
					break;
			case SDLK_g:	m_clrbg.r = m_clrbg.g = m_clrbg.b = 0;
					m_clrfg.r = 0; 	  m_clrbr.r = 0;
					m_clrfg.g = 0xc0; m_clrbr.g = 0xff;
					m_clrfg.b = 0;	  m_clrbr.b = 0;
					break;
			case SDLK_w:	m_clrbg.r = m_clrbg.g = m_clrbg.b = 0;
					m_clrfg.r = 0xc0;  m_clrbr.r = 0xff;
					m_clrfg.g = 0xc0;  m_clrbr.g = 0xff;
					m_clrfg.b = 0xc0;  m_clrbr.b = 0xff;
					break;
			case SDLK_a:	m_clrbg.r = m_clrbg.g = m_clrbg.b = 0;
					m_clrfg.r = 0xc0; m_clrbr.r = 0xff;
					m_clrfg.g = 0x80; m_clrbr.g = 0xa0;
					m_clrfg.b = 0;	  m_clrbr.b = 0;
					break;
			case SDLK_ESCAPE: return UIE_CONTINUE;
			default:	break;
		}
		m_bg = SDL_MapRGB(vidSurface->format, m_clrbg.r, m_clrbg.g, m_clrbg.b);
		m_fg = SDL_MapRGB(vidSurface->format, m_clrfg.r, m_clrfg.g, m_clrfg.b);
		m_bright = SDL_MapRGB(vidSurface->format, m_clrbr.r, m_clrbr.g, m_clrbr.b);
	}
	return UIE_CONTINUE;
}

