/*************************************************************************
 
    Quick And Dirty Apricot Emulator
    Copyright (C) 2005  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "qdae.hxx"
#include "device.hxx"
#include "z80_ctc.hxx"


//#define CTC_TRACE(s) do { printf("Z80 CTC: "); printf s; putchar('\n'); \
//	fflush(stdout);} while(0)

#define CTC_TRACE(s)
//
// A fudge factor. It's the number of CTC clock ticks to simulate per 8086 
// instruction cycle.
//
#define TICK_DIVIDER 6

Z80_CTC::Z80_CTC(PC_ENV *host) : Device("z80ctc")
{
	pc_env = host;
	raising = 0;
	irq_base = 0;
	last_z80_op = 0;
	for (int n = 0; n < 4; n++)
	{
		expect_timer_word[n] = 0;
		timer_word[n] = 0;
		countdown[n] = 0;
		channel_word[n] = 0;
	}
	reset();
}

Z80_CTC::~Z80_CTC()
{

}


void Z80_CTC::opcode(u_int8 value)
{
	if (last_z80_op == 0xED && value == 0x4B)
	{
		raising = 0;
	}
	last_z80_op = value;
}

void Z80_CTC::write(u_int16 channel, u_int8 value)
{
	if (expect_timer_word[channel])
	{
		CTC_TRACE(("Channel %d timer word set to 0x%02x", channel, value));
		timer_word[channel] = value;
		set_countdown(channel);
		CTC_TRACE(("Channel %d countdown set to 0x%04x", channel, 
					countdown[channel]));
		expect_timer_word[channel] = 0;
		return;
	}

	if (value & 1)	/* Set timer interval */
	{
		channel_word[channel] = value;
		if (value & 2) reset();
		if (value & 4) expect_timer_word[channel] = 1;

		CTC_TRACE(("Channel %d programmed with 0x%02x", channel, value));
	}
	else
	{
		if (channel == 0) 
		{
			irq_base = value & 0xF8;
			CTC_TRACE(("Interrupt base set to 0x%02x", irq_base));
		}
	}	
}	

u_int8 Z80_CTC::read(u_int16 channel)
{
	u_int8 value;

	if (channel_word[channel] & 0x20) 
		value = countdown[channel] * TICK_DIVIDER / 256;
	else	value = countdown[channel] * TICK_DIVIDER / 16;


	CTC_TRACE(("Read channel %d returns 0x%02x", channel, value));

	return value;
}

void Z80_CTC::reset()
{
	raising = 0;
}

u_int8 Z80_CTC::ackInterrupt()
{
	raising = 0;
	return last_irq;
}

void Z80_CTC::raise(int channel)
{
/* Raise the interrupt anyway. If the CPU doesn't listen, that's its
 * lookout. */
	last_irq = (raising - 1) * 2 + irq_base;
	pc_env->i86_intr_raise(this);
}


u_int8 Z80_CTC::tick()
{
	int channel, interrupted;

	interrupted = 0;
	for (channel = 0; channel < 4; channel++)
	{
/* Rule of thumb: Count down 1 tick per CPU cycle */
		if (countdown[channel])
		{
			countdown[channel]--;
			if (countdown[channel] == 0)
			{
				set_countdown(channel);
			
				if (!interrupted && (channel_word[channel] & 0x80))
				{
					interrupted = 1;
					raising = channel + 1;
				}
			}
		}
	}
	/* If we decide to interrupt, say so */
	if (raising)
	{
		raise(raising - 1);
	}
	return 0;
}

void Z80_CTC::set_countdown(int channel)
{
	if (channel_word[channel] & 0x20) 
		countdown[channel] = timer_word[channel] * 256 + 255;
	else	countdown[channel] = timer_word[channel] * 16 + 15;

	countdown[channel] /= TICK_DIVIDER;

}


bool Z80_CTC::parseNode(xmlDoc *doc, xmlNs *ns, xmlNode *node)
{
	return true;
}

bool Z80_CTC::storeNode(xmlDoc *doc, xmlNs *ns, xmlNode *node)
{
	return true;
}

